structure Logic = struct
  (* This type is empty. By definition, a value of type void can never be constructed. *)
  (* SML doesn't let us define a type with no constructors, so this is the best we can get. *)
  datatype void = Void of void
  (* Believe it or not this function is provably total *)
  fun abort (Void v) = abort v
  
  (* Commonly used hypothesis names are "u", "v", and "w" *)
  type hypothesis = string

  (*
   * A proof tree is either:
   * - A proof of a judgment using an inference rule and premises
   * - An invocation of a theorem to show a judgment
   * - A referece to a judgment which is considered given
   *)
  datatype ('given,'theorem,'rule,'judgment) abstractprooftree
    = ProofTree of ('given,'theorem,'rule,'judgment) abstractprooftree list * 'rule    * 'judgment
    | Theorem   of ('given,'theorem,'rule,'judgment) abstractprooftree list * 'theorem * 'judgment
    | Given     of  'given * 'judgment

  (*
   * In logic, a metavariable is a symbol which belongs to a metalanguage and
   * stands for elements of some object language.
   *
   * We haven't defined what atomic facts our logic is talking about.
   * That means we need to use "metavariables". You don't need more
   * than five of these in the homeworks we give you.
   * https://en.wikipedia.org/wiki/Metavariable
   *)
  datatype atom = AA | AB | AC | AD | AE

  (* 
   * Macros for writing proofs
   * Single lines are for inference rules
   * Double lines are for invoking theorems
   * All line lengths are identical
   * These will be infixed for you in each HW's starter code
   *)
  fun -- (proofs, rule) judgment = ProofTree (proofs,rule,judgment)
  fun == (proofs, theorem) judgment = Theorem (proofs,theorem,judgment)
  val (---,==========) = (--,==)
  val (----,=========) = (--,==)
  val (-----,========) = (--,==)
  val (------,=======) = (--,==)
  val (-------,======) = (--,==)
  val (--------,=====) = (--,==)
  val (---------,====) = (--,==)
  val (----------,===) = (--,==)
end