structure PrettyPrinter : sig
  (* val pretty_print_prooftree : ('given -> string) * ('theorem -> string) * ('rule -> string) * ('judgment -> string) -> ('given,'theorem,'rule,'judgment) Logic.abstractprooftree -> unit *)
  val pretty_print_proof : ('rule -> string) * ('judgment -> string) -> ('given,'theorem,'rule,'judgment) Logic.abstractprooftree -> unit
end = struct
  type ('given,'theorem,'rule,'judgment) printctx = ('given -> string) * ('theorem -> string) * ('rule -> string) * ('judgment -> string)
  open Logic
  fun spaces i = String.implode (List.tabulate (i,(fn _ => #" ")))
  fun pad str width =
    let
      val missing = (width - String.size str)
      val left = missing div 2
      val right = missing - left
    in
      spaces left ^ str ^ spaces right
    end
  fun bar i = String.implode (List.tabulate (i,(fn _ => #"-")))
  fun write_rule rule size = bar (size - String.size rule - 1) ^ rule ^ " "
  fun join_trees (size1,size2) =
    fn([],[]) => []
    | (x::xs,[]) => (x ^ spaces size2) :: join_trees (size1,size2) (xs,[])
    | ([],y::ys) => (spaces size1 ^ y) :: join_trees (size1,size2) ([],ys)
    | (x::xs,y::ys) => (x ^ y) :: join_trees (size1,size2) (xs,ys)
  fun proof_to_tree (ctx as (given_to_string,theorem_to_string,rule_to_string,judgment_to_string)) ((ProofTree (L,r,j))) =
    let
      val L = map (proof_to_tree ctx) L
      (* val (size,over) = foldl (fn (size2,_,over2) => fn (size1,_,_) => (size1+size2,over2)) (0,"") L *)
      val (size, str, over) = foldl (fn ((size2,str2,over2),(size1,str1,over1)) => (size1+size2,join_trees (size1,size2) (str1,str2),over2)) (0,[],0) L
      (* val (L_len,buffer) = foldr (fn {sub;j;rule;width;over} => fn (i,f) => (width + i + f,over)) (0,0) *)
      val r = " " ^ rule_to_string r ^ " "
      val r_len = String.size r
      val overfill = r_len - over
      val (size,str,over,r) = if overfill > 0 then let val buff = spaces overfill in (size+overfill,map (fn s => s ^ buff) str,over + overfill,r) end else (size,str,over,r ^ spaces (~overfill))
      val j = " " ^ judgment_to_string j ^ " "
      val j_len = String.size j
      val overfill = j_len + over - size
      val (size,str,j) = if overfill > 0 then let val size = size + overfill in (size,map (fn s => (if String.isPrefix "-" s then bar else spaces) overfill ^ s) str,j ^ spaces over) end else (size,str,pad j (size - over) ^ spaces over)
      val j_len = String.size j
      val r = bar (j_len-over) ^ r
    in
      (size,j::r::str,over)
    end
  | proof_to_tree ctx (Given _ | Theorem _) = raise Fail "Unimplemented printing of theorems and givens"
  fun pretty_print_prooftree ctx p = (print "\n"; foldr (fn (s,()) => print (s ^ "\n")) () (#2 (proof_to_tree ctx p)); print "\n")
  fun pretty_print_proof (r2s,j2s) = pretty_print_prooftree (Logic.abort,Logic.abort,r2s,j2s)
end